#include "common.h"
#include "snmppd.h"
#include "control.h"
#include "queue.h"
#include "dbug.h"
#include <splitter.h>

#ifdef USE_THREADS
#include <pthread.h>
#include <threads.h>
#endif /* USE_THREADS */

#include <asm/poll.h>
#include <sys/types.h>
#include <signal.h>
#include <errno.h>

#ifdef MALLOC_DEBUG
#include "rmalloc/rmalloc.h"
#endif


stats_t stats = { 0, 0, 0, 0, 0, 0, 0, 0.0 };
struct snmp_session session;
#ifdef USE_THREADS
pthread_mutex_t mem_mutex;
work_queue wq;
cleanup_queue cq;
#endif
config_t config;

void cleanup()
{
#ifdef USE_THREADS
    wnode *mywork;
    
    if (config.debug) log_debug("deactivating work queue...");
    control_deactivate(&wq.control);
    while ((mywork = (wnode *) queue_get(&wq.work)) != NULL) {
	pthread_mutex_lock(&mem_mutex);
	free(mywork);
	pthread_mutex_unlock(&mem_mutex);
    }
    join_threads();
    cleanup_structs();
    pthread_mutex_destroy(&mem_mutex);
#endif /* USE_THREADS */
    unlink(config.pid_file);
    exit(0);
}

int main(int argc, char *argv[])
{
    int ch;
#ifdef USE_THREADS
    wnode *mywork;
#else
    int cantFork = 0;
    int ppid;
#endif /* USE_THREADS */
    char buf[MAX_SNMPPD_OID_LEN];
//    char buf2[1024];
    struct sockaddr_in addr;
    int ssockfd, csockfd;
    int clen;
    struct sockaddr_in caddr;
    int ctot, i, n, nready;
    target_t *entry;
    int one = 1;
    int cycle = 0;
    struct pollfd pollfds[FD_SETSIZE];
    
    config.nofork = 0;
    config.debug = 0;
    config.verbose = 0;
    sprintf(config.pid_file, "/var/run/snmppd.pid");
    while ((ch = getopt(argc, argv, "fvhd")) != EOF) {
	switch ((char) ch) {
	case 'f':
	    config.nofork = 1;
	    break;
	case 'v':
	    config.verbose = 1;
	    break;
	case 'h':
	    usage(argv[0]);
	    break;
	case 'd':
	    config.verbose = 1;
	    config.debug = 1;
	    config.nofork = 1;
	    break;
	default:
	    break;
	}
    }

    memset(buf, 0 , MAX_SNMPPD_OID_LEN);
    if ((!config.debug) && (!config.nofork)) {
	openlog("snmppd", LOG_PID, LOG_DAEMON);
    }
    if (!config.nofork) {
	if (fork() != 0) {
	    exit(0);
	}
	setsid();
    }
    if (write_pid())
	exit(1);

    snmppd_snmp_init();

#ifdef USE_THREADS
    initialize_structs();
    pthread_mutex_init(&mem_mutex, NULL);

    if (create_threads()) {
	log_error("Error starting threads... cleaning up.");
	join_threads();
	dabort();
    }
#endif /* USE_THREADS */
    signal(SIGINT, cleanup);
    signal(SIGTERM, cleanup);


    if ((ssockfd=socket(PF_INET, SOCK_STREAM, 0)) == -1){
	log_error ("Can't create socket: %s", strerror(errno));
	exit(-1);
    }
    if (setsockopt(ssockfd, SOL_SOCKET, SO_REUSEADDR,
        (char *)&one, sizeof(int)) < 0) {
	log_error ("Can't setsockopt(SO_REUSEADDR)");
	close(ssockfd);
    }    
    memset(&addr, 0 ,sizeof(addr));
    addr.sin_addr.s_addr = htonl (INADDR_LOOPBACK);
    addr.sin_family = AF_INET;
    addr.sin_port = htons(SNMPPD_PORT);
    
    if (bind(ssockfd, (struct sockaddr *) &addr, sizeof(addr)) == -1) {
	log_error ("Can't bind: %s", strerror(errno));
	exit(-1);
    }
    if (listen(ssockfd, 5) == -1){
	log_error ("Can't listen: %s", strerror(errno));
	exit(-1);
    }
    /* Initialize pollfds array with server, client info */
    ctot = 0;			// Just the server initially, no clients
    pollfds[0].fd = ssockfd;
    pollfds[0].events = POLLRDNORM;
    for (i = 1; i < FD_SETSIZE; i++)
	pollfds[i].fd = -1;

    while (1){
	cycle++;
	if (config.debug) log_debug ("server waiting");
	nready = poll (pollfds, ctot + 1, -1);

      /* Server, need to add new client */
	if (pollfds[0].revents & POLLRDNORM){
	    if (config.debug) log_debug ("new client: %d", ctot + 1);
	    clen = sizeof (caddr);
	    csockfd = accept (ssockfd, (struct sockaddr *) &caddr, &clen);
	    for (i = 1; i < FD_SETSIZE; i++)
		if (pollfds[i].fd < 0){
			pollfds[i].fd = csockfd;
			break;
		}

	    if (i == FD_SETSIZE){
		log_error("too many clients");
		exit (1);
	    }

	    pollfds[i].events = POLLRDNORM;
	    if (i > ctot)
		ctot = i;

	    if (--nready <= 0)
		continue;
	}

      /* Now service clients in ascending order */
	for (i = 1; i <= ctot; i++){
	    if ((csockfd = pollfds[i].fd) < 0)
		continue;

	    if (config.debug) log_debug("fd %d, i %d: ", csockfd, i);
	    if (pollfds[i].revents & (POLLRDNORM | POLLERR)){
		if (config.debug) log_debug ("cycle %d, fd %d: Reading", cycle, csockfd);
		if ((n = read (csockfd, &buf, MAX_SNMPPD_OID_LEN)) < 0){
		    if (config.debug) log_debug("fd %d: Error on socket (%s)", csockfd, strerror(errno));
		    close (csockfd);
		    pollfds[i].fd = -1;
		    continue;
		}
		else if (n == 0){
		    close (csockfd);
		    if (config.debug) log_debug ("fd %d: Removing client", csockfd);
		    pollfds[i].fd = -1;
		    continue;
		}
		else {
		    char **params = NULL;
		    int j = 0;
		    char *p = NULL;
		    object_t *obj_tail = NULL;
		    
		    buf[n] = '\0';
		    while ((p=strchr(buf, '\n')) != NULL){
			*p='\0';
		    }
		    while ((p=strchr(buf, '\r')) != NULL){
			*p='\0';
		    }
		    if (config.debug) log_debug ("cycle %d, fd %d: Serving client",  cycle, csockfd);
		    if (config.debug) log_debug ("cycle %d, fd %d: Request was\n|%s|\n(%d bytes)",  cycle, csockfd, buf, n);
//		    log_info ("fd %d: Request: %s", csockfd, buf);
//		    pthread_mutex_lock(&mem_mutex);
#ifdef USE_THREADS
		    pthread_mutex_lock(&mem_mutex);
#endif /* USE_THREADS */
		    buf[n]='\0';
		    params = split(buf, " ", SPLIT_INFINITE, SPLIT_REPEAT);
		    j = 3;
		    while (params[j] != NULL){
			if (config.debug) log_debug("cycle %d, fd %d: Got param %d: %s", cycle, csockfd, j - 3, params[j]);
			j++;
		    }
		    entry = malloc(sizeof(target_t));
		    memset(entry, 0, sizeof(target_t));
		    entry->sock = csockfd;
		    entry->hostname = strdup(params[1]);
		    entry->community = strdup(params[2]);
//		    entry->objects = NULL;
#ifdef USE_THREADS
		    pthread_mutex_unlock(&mem_mutex);
#endif /* USE_THREADS */
		    j=3;
		    while ((params[j] != NULL) && (params[j][0] != '\0')){
			object_t *tmp_obj = NULL;
#ifdef USE_THREADS
			pthread_mutex_lock(&mem_mutex);
#endif /* USE_THREADS */
			tmp_obj = (object_t *)malloc(sizeof(object_t));
			memset(tmp_obj, 0, sizeof(object_t));
			
			if (config.debug) log_debug("fd %d: Adding object id to target struct: %s", csockfd, params[j]);
			strncpy(tmp_obj->objid, params[j], MAX_SNMPPD_OID_LEN);
			if (config.debug) log_debug("fd %d: Added object id to target struct: %s", csockfd, tmp_obj->objid);
			tmp_obj->next = NULL;
			if (!obj_tail){
			    entry->objects = tmp_obj;
			    obj_tail = tmp_obj;
			}else{
			    obj_tail->next = tmp_obj;
			    obj_tail = obj_tail->next;
			}
			j++;
#ifdef USE_THREADS
			pthread_mutex_unlock(&mem_mutex);
#endif /* USE_THREADS */
		    }
		    
#ifdef USE_THREADS
		    pthread_mutex_lock(&mem_mutex);
#endif /* USE_THREADS */
		    split_free(params);
#ifdef USE_THREADS
		    pthread_mutex_unlock(&mem_mutex);
		    pthread_mutex_lock(&wq.control.mutex);
		    pthread_mutex_lock(&mem_mutex);
		    mywork = malloc(sizeof(wnode));
		    memset(mywork, 0, sizeof(wnode));
		    pthread_mutex_unlock(&mem_mutex);
		    if (!mywork) {
			log_error("ouch! can't malloc!");
			break;
		    }
		    //c++;
		    mywork->next=NULL;
		    mywork->entry = entry;
		    queue_put(&wq.work, (node *) mywork);
		    if (config.debug) log_debug("%d elements in queue after put", queue_size(&wq.work));
		    pthread_mutex_unlock(&wq.control.mutex);
		    if (config.debug) log_debug("Sending signal to workers");
		    pthread_cond_signal(&wq.control.cond);
#else /* No-threads work handling */
		    if ((ppid = fork()) < 0){
			log_error("Can't fork");
		    }
		    
		    if (config.debug) log_debug("cycle %d: Process %d, ppid == %d ", cycle, getpid(), ppid);
		    if (ppid == 0){ /* Child */
#ifdef USE_SETSID
			setsid();
#else
			if (fork()==0){
#endif
			    if (config.debug) log_debug("cycle %d: Child %d detached, running child_main()", cycle, getpid());
			    child_main(entry);
#ifdef UDE_SETSID
#else
			}
#endif
			if (config.debug) log_debug("cycle %d: Child %d exiting", cycle, getpid());
			exit(0);
		    }else{ /* Parent */
			wait(NULL);
			if (config.debug) log_debug("cycle %d: Parent continuing after running child %d", cycle, ppid);
			free_entry(entry);
			close (csockfd);
			if (config.debug) log_debug("cycle %d, fd %d: Removing client", cycle, csockfd);
			pollfds[i].fd = -1;
			continue;
		    }
#endif /* USE_THREADS */
		}

		if (--nready <= 0)
			break;

	    }

	}
///
///
#ifdef MALLOC_DEBUG
	RM_STAT;
#endif		
    }
    if (config.debug) log_debug ("Parent exiting");
    cleanup();
}



#ifdef USE_THREADS
void *thread_poll(void *myarg)
{
    wnode *mywork;
    cnode *mynode;
    char *buf = NULL;
    int result;
    unsigned long long counter = 0ULL;

    sigset_t set;
//    buf=(char **)malloc(sizeof(char**));
    /*
     *      Block SIGHUP handling for the child threads.
     *
     *      This ensures that only the main server thread will
     *      process HUP signals.
     *
     *      If we don't have sigprocmask, then it shouldn't be
     *      a problem, either, as the sig_hup handler should check
     *      for this condition.
     */
    sigemptyset(&set);
    sigaddset(&set, SIGHUP);
    sigaddset(&set, SIGINT);
    sigaddset(&set, SIGQUIT);
    sigaddset(&set, SIGPIPE);
    pthread_sigmask(SIG_BLOCK, &set, NULL);

    mynode = (cnode *) myarg;

    pthread_mutex_lock(&wq.control.mutex);
    while (wq.control.active) {
	while (wq.work.head == NULL && wq.control.active) {
	    if (config.debug) log_debug("Thread %d is waiting for control condition",
		      mynode->threadnum);
	    pthread_cond_wait(&wq.control.cond, &wq.control.mutex);
	}
	if (!wq.control.active)
	    break;
	//we got something!
	mywork = (wnode *) queue_get(&wq.work);
	if (config.debug) log_debug("%d elements remains in queue after get", queue_size(&wq.work));
/*	if (!mywork){
	    continue;
	}*/
	pthread_mutex_unlock(&wq.control.mutex);
	if (config.debug) log_debug("Thread number %d got job", mynode->threadnum);
	//perform processing...
	mywork->entry->tid = mynode->threadnum;
#if 1
	if (!mywork->entry){
	    if (config.debug) log_debug("thread %d, fd %d: Error getting result (entry is NULL)", mywork->entry->tid, mywork->entry->sock);
	    write(mywork->entry->sock, "ERROR\n", 6);
	    pthread_mutex_lock(&wq.control.mutex);
	    continue;
	}
	if (!mywork->entry->objects){
	    if (config.debug) log_debug("thread %d, fd %d: Error getting result (entry->objects is NULL)", mywork->entry->tid, mywork->entry->sock);
	    write(mywork->entry->sock, "ERROR\n", 6);
	    pthread_mutex_lock(&wq.control.mutex);
	    continue;
	}
	if (!mywork->entry->objects->objid){
	    if (config.debug) log_debug("thread %d, fd %d: Error getting result (entry->objects->objid is NULL)", mywork->entry->tid, mywork->entry->sock);
	    write(mywork->entry->sock, "ERROR\n", 6);
	    pthread_mutex_lock(&wq.control.mutex);
	    continue;
	}
	if (!mywork->entry->hostname){
	    if (config.debug) log_debug("thread %d, fd %d: Error getting result (entry->hostname is NULL)", mywork->entry->tid, mywork->entry->sock);
	    write(mywork->entry->sock, "ERROR\n", 6);
	    pthread_mutex_lock(&wq.control.mutex);
	    continue;
	}
	if (!mywork->entry->community){
	    if (config.debug) log_debug("thread %d, fd %d: Error getting result (entry->community is NULL)", mywork->entry->tid, mywork->entry->sock);
	    write(mywork->entry->sock, "ERROR\n", 6);
	    pthread_mutex_lock(&wq.control.mutex);
	    continue;
	}
#endif
	if (config.debug) log_debug("Thread number %d processing job %s@%s",
		  mynode->threadnum, mywork->entry->objects->objid,
		  mywork->entry->hostname);
	buf = (char *)malloc(8192);
	memset(buf, 0, 8192);
	result = poll_target(mywork->entry, &buf, 8192);
	if (result != 0){
	    if (config.debug) log_debug("thread %d, fd %d: Error getting result", mywork->entry->tid, mywork->entry->sock);
	    write(mywork->entry->sock, "ERROR\n", 6);
	}else{
	    if (config.debug) log_debug("thread %d, fd %d: Got result\n|%s|", mywork->entry->tid, mywork->entry->sock, buf);
	    write(mywork->entry->sock, buf, strlen(buf));
//	    write(mywork->entry->sock, "\n", 1);
	}
	counter++;
	pthread_mutex_lock(&mem_mutex);
	while (mywork->entry->objects){
	    object_t *tmp_obj = mywork->entry->objects;
	    mywork->entry->objects = mywork->entry->objects->next;
//	    if (tmp_objobjid) free(tmp_obj->objid);
	    free(tmp_obj);
	}
	free(buf);
	free(mywork->entry->hostname);
	free(mywork->entry->community);
	free(mywork->entry);
	free(mywork);
	pthread_mutex_unlock(&mem_mutex);
	pthread_mutex_lock(&wq.control.mutex);
    }
    pthread_mutex_unlock(&wq.control.mutex);
    pthread_mutex_lock(&cq.control.mutex);
    queue_put(&cq.cleanup, (node *) mynode);
    pthread_mutex_unlock(&cq.control.mutex);
    pthread_cond_signal(&cq.control.cond);
    if (config.verbose)
    log_info("thread %d shutting down after %llu polls.",
	     mynode->threadnum, counter);
    if (config.debug) log_debug("thread %d shutting down...", mynode->threadnum);
    return NULL;


}
#else /* USE_THREADS */

void child_main(target_t *entry){
    char *buf = NULL;
    int result;

    buf = (char *)malloc(8192);
    memset(buf, 0, 8192);
    result = poll_target(entry, &buf, 8192);
    if (result != 0){
        if (config.debug) log_debug("child %d, fd %d: Error getting result", getpid(), entry->sock);
	    write(entry->sock, "ERROR\n", 6);
    }else{
	if (config.debug) log_debug("child %d, fd %d: Got result\n|%s|", getpid(), entry->sock, buf);
//	log_info("child %d, fd %d: Got result: %s", getpid(), entry->sock, buf);
	write(entry->sock, buf, strlen(buf));
//	write(mywork->entry->sock, "\n", 1);
    }
    free(buf);
    free_entry(entry);
}
#endif /* USE_THREADS */

void free_entry(target_t *entry){
    while (entry->objects){
        object_t *tmp_obj = entry->objects;
        entry->objects = entry->objects->next;
	free(tmp_obj);
    }
    free(entry->hostname);
    free(entry->community);
    free(entry);
}

void usage(char *prog)
{
    printf("%s v%s\n", PACKAGE, VERSION);
    printf("Usage: %s [-d][-f] \n", prog);
    printf("       -d: enable debug; -f: stay foreground\n");
    exit(-1);
}
