#include "common.h"
#include "snmppd.h"

#include "asn1.h"
#include "snmp_api.h"
#include "snmp_impl.h"
#include "snmp_client.h"
#include "mib.h"
#include "snmp.h"
#ifdef USE_THREADS
#include <pthread.h>
#endif /* USE_THREADS */

#ifdef MALLOC_DEBUG
#include "rmalloc/rmalloc.h"
#endif


//pthread_mutex_t snmp_mutex;


/* Initialize the snmp_session struct and set defaults */
void snmppd_snmp_init()
{
    setenv("MIBS", "all", 1);
    init_snmp("SNMPPD");
    snmp_sess_init(&session);
    session.version = SNMP_VERSION_1;
    session.retries = SNMP_DEFAULT_RETRIES;
    session.timeout = 8 * 1000000L;
    session.remote_port = SNMP_DEFAULT_REMPORT;
    snmp_out_toggle_options("Qu");
}

/* Perform the actual SNMP calls and return the retrieved SNMP value */
int poll_target(target_t * target, char **buf, int buf_size)
{
    oid name[MAX_OID_LEN];
    size_t name_length = MAX_OID_LEN;
    struct snmp_session *ss = NULL;
    void *sessp = NULL;
    struct snmp_pdu *pdu = NULL, *response = NULL;
    struct variable_list *vars = NULL;
//    long long poll_val = 0;
    int status;
//    char result[256];
    object_t *object = NULL;
    int objects = 0;
    int failures = 0;

#ifdef USE_THREADS
    pthread_mutex_lock(&mem_mutex);
#endif /* USE_THREADS */
    session.peername = target->hostname;
    session.community = (u_char *) target->community;
    session.community_len = strlen((char *) target->community);

    sessp = snmp_sess_open(&session);
#ifdef USE_THREADS
    pthread_mutex_unlock(&mem_mutex);
#endif /* USE_THREADS */

    if (sessp == NULL) {
//	char *err = NULL;
//	snmp_sess_error(&session, NULL, NULL, &err);
#ifdef USE_THREADS
	log_error("thread %d, fd %d: Couldn't open snmp descriptor when polling %s@%s", target->tid, target->sock, target->objects->objid, target->hostname);
#else
	log_error("child %d, fd %d: Couldn't open snmp descriptor when polling %s@%s", getpid(), target->sock, target->objects->objid, target->hostname);
#endif /* USE_THREADS */
//	if(err) free(err);
	return (-2);
    }
#ifdef USE_THREADS
    pthread_mutex_lock(&mem_mutex);
#endif /* USE_THREADS */
    ss = snmp_sess_session(sessp);
    pdu = snmp_pdu_create(SNMP_MSG_GET);
#ifdef USE_THREADS
    pthread_mutex_unlock(&mem_mutex);
#endif /* USE_THREADS */

    object = target->objects;
    while (object){
	name_length = MAX_OID_LEN;
#ifdef USE_THREADS
        pthread_mutex_lock(&mem_mutex);
#endif /* USE_THREADS */
#if 1
	if (!snmp_parse_oid(object->objid, name, &name_length)) {
#else
	if (!read_objid(object->objid, name, &name_length)) {
#endif
#ifdef USE_THREADS
	    log_error("thread %d, fd %d: Invalid object identifier on %s: [%s]\n", target->tid, target->sock, target->hostname, target->objects->objid);
#else
	    log_error("child %d, fd %d: Invalid object identifier on %s: [%s]\n", getpid(), target->sock, target->hostname, target->objects->objid);
#endif /* USE_THREADS */
	    snmp_perror(object->objid);
	    failures++;
	}else{
	    snmp_add_null_var(pdu, name, name_length);
	}
	object=object->next;
	objects++;
#ifdef USE_THREADS
	pthread_mutex_unlock(&mem_mutex);
#endif /* USE_THREADS */
    }
#ifdef USE_THREADS
    if (config.debug) log_debug("thread %d, fd %d: %d objects id's parsed", target->tid, target->sock, objects);
#else
    if (config.debug) log_debug("child %d, fd %d: %d objects id's parsed", getpid(), target->sock, objects);
#endif /* USE_THREADS */
    if (failures > 0){
#ifdef USE_THREADS
	pthread_mutex_lock(&mem_mutex);
#endif /* USE_THREADS */
	snmp_sess_close(sessp);
#ifdef USE_THREADS
	pthread_mutex_unlock(&mem_mutex);
	log_error("thread %d, fd %d: Found failures when polling %s@%s", target->tid, target->sock, target->objects->objid, target->hostname);
#else
	log_error("child %d, fd %d: Found failures when polling %s@%s", getpid(), target->sock, target->objects->objid, target->hostname);
#endif /* USE_THREADS */
	return (-2);
    }
//    snmp_add_null_var(pdu, name, name_length);
    status = snmp_sess_synch_response(sessp, pdu, &response);

    /* No Response */
    if (status == STAT_TIMEOUT) {
#ifdef USE_THREADS
	pthread_mutex_lock(&mem_mutex);
#endif /* USE_THREADS */
	snmp_sess_close(sessp);
#ifdef USE_THREADS
	pthread_mutex_unlock(&mem_mutex);
#endif /* USE_THREADS */
	return (-1);
    }
    /* Error */
    else if (status != STAT_SUCCESS) {
#ifdef USE_THREADS
	pthread_mutex_lock(&mem_mutex);
#endif /* USE_THREADS */
	snmp_sess_close(sessp);
#ifdef USE_THREADS
	pthread_mutex_unlock(&mem_mutex);
#endif /* USE_THREADS */
	return (-2);
    }
    /* Received a response */
    else {
	if (response->errstat != SNMP_ERR_NOERROR) {
	    /* Error with the packet. */
#ifdef USE_THREADS
	    log_error("thread %d, fd %d: SNMP Object Error: %s\n", target->tid, target->sock, 
		      snmp_errstring(response->errstat));
	    log_error("thread %d, fd %d: [host=%s] [oid=%s]\n", target->tid, target->sock, target->hostname,
		      target->objects->objid);
	    pthread_mutex_lock(&mem_mutex);
#else
	    log_error("child %d, fd %d: SNMP Object Error: %s\n", getpid(), target->sock, 
		      snmp_errstring(response->errstat));
	    log_error("child %d, fd %d: [host=%s] [oid=%s]\n", getpid(), target->sock, target->hostname,
		      target->objects->objid);
#endif /* USE_THREADS */
	    snmp_free_pdu(response);
	    snmp_sess_close(sessp);
#ifdef USE_THREADS
	    pthread_mutex_unlock(&mem_mutex);
#endif /* USE_THREADS */
	    return (-2);
	} else {
	    int offset = 0;
#ifdef USE_THREADS
	    if (config.debug) log_debug("thread %d, fd %d: Received SNMP response from %s", target->tid, target->sock, target->hostname);
#else
	    if (config.debug) log_debug("child %d, fd %d: Received SNMP response from %s", getpid(), target->sock, target->hostname);
#endif /* USE_THREADS */
	    vars = response->variables;
	    while (vars){
	    	    offset+=snprint_variable(*buf+offset, buf_size-offset, vars->name, vars->name_length, vars);
		    strncpy((*buf)+offset, "\n", 1);
		    offset++;
		    vars = vars->next_variable;
	    }
#ifdef USE_THREADS
	    if (config.debug) log_debug("thread %d, fd %d: Returning to client: \n%s", target->tid, target->sock, *buf);
	    pthread_mutex_lock(&mem_mutex);
#else
	    if (config.debug) log_debug("child %d, fd %d: Returning to client: \n%s", getpid(), target->sock, *buf);
#endif /* USE_THREADS */
	    snmp_free_pdu(response);
	    snmp_sess_close(sessp);
#ifdef USE_THREADS
	    pthread_mutex_unlock(&mem_mutex);
#endif /* USE_THREADS */
	    return (0);
	}
    }
}
